@@ -1,5 +1,3 @@
-#!/usr/bin/env python
-
 #
 # Copyright (C) 2005 Red Hat, Inc.
 #
@@ -21,12 +19,12 @@ import os
 import os.path
 
 PACKAGE    = "sabayon"
-VERSION    = "2.12.3"
-LOCALEDIR = "/opt/gnome2/share/locale"
-DATADIR    = "/opt/gnome2/share"
-SYSCONFDIR = "/opt/gnome2/etc"
-LIBEXECDIR = "/opt/gnome2/libexec"
-SBINDIR    = "/opt/gnome2/sbin"
+VERSION    = "2.12.4"
+LOCALEDIR = "/gnome/head/INSTALL/share/locale"
+DATADIR    = "/gnome/head/INSTALL/share"
+SYSCONFDIR = "/gnome/head/INSTALL/etc"
+LIBEXECDIR = "/gnome/head/INSTALL/libexec"
+SBINDIR    = "/gnome/head/INSTALL/sbin"
 
 GLADEDIR    = os.path.join (DATADIR, "sabayon", "glade")
 PROFILESDIR = os.path.join (SYSCONFDIR, "desktop-profiles")
@@ -53,8 +51,8 @@ DEFAULT_SHELL = "/bin/bash"
 NOLOGIN_SHELL = "/sbin/nologin"
 SKEL_HOMEDIR  = "/etc/skel"
 
-XNEST_ARGV   = [ "/usr/X11R6/bin/Xnest", "-terminate", "-audit", "0", "-nolisten", "tcp" ]
-SESSION_ARGV = [ "/etc/X11/xdm/Xsession", "gnome" ]
+XNEST_ARGV   = [ "/usr/bin/Xnest", "-terminate", "-audit", "0", "-nolisten", "tcp" ]
+SESSION_ARGV = [ "/etc/X11/xinit/Xsession", "gnome" ]
 USERMOD_ARGV = [ "/usr/sbin/usermod" ]
 
 PANEL_KEY_BASE = "/apps/panel"
@@ -75,7 +73,8 @@ PASSTHROUGH_ENVIRONMENT = [
     "LC_MONETARY",
     "LC_MESSAGES",
     "METACITY_VERBOSE",
-    "METACITY_USE_LOGFILE"
+    "METACITY_USE_LOGFILE",
+    "PYTHONPATH"
 ]
 
 #
