@@ -1,5 +1,3 @@
-#!/usr/bin/env python
-
 #
 # Copyright (C) 2005 Red Hat, Inc.
 #
@@ -44,6 +42,7 @@ class FilesChange (userprofile.ProfileChange):
         userprofile.ProfileChange.__init__ (self, source)
         self.rel_path = rel_path
         self.event    = event
+        self.created = self.event == dirmonitor.CREATED
 
         assert self.event == dirmonitor.CREATED or \
                self.event == dirmonitor.DELETED or \
@@ -52,6 +51,10 @@ class FilesChange (userprofile.ProfileChange):
     def get_id (self):
         return self.rel_path
 
+    def merge_old_change (self, old_change):
+        self.created = old_change.created
+        return self.event == dirmonitor.DELETED and self.created
+
     def get_ignore_default (self):
         # Ignore backup files by default
         if self.rel_path[-1] == "~":
@@ -79,7 +82,7 @@ class FilesSource (userprofile.ProfileSource):
         self.monitor.set_files_to_ignore (FILES_TO_IGNORE)
 
     def __handle_monitor_event (self, path, event):
-        if os.path.isfile (path):
+        if event == dirmonitor.DELETED or os.path.isfile (path):
             # FIXME: sanity check input (e.g. is change actually in homedir/?)
             rel_path = path[len (self.home_dir):].lstrip ("/")
             dprint ("Emitting event '%s' on file '%s'",
