@@ -1,5 +1,3 @@
-#!/usr/bin/env python
-
 #
 # Copyright (C) 2005 Red Hat, Inc.
 #
@@ -23,17 +21,20 @@ import exceptions, sys, os.path, ConfigParser, re, cPickle
 import tempfile, types
 import traceback
 import errno
+import time
 
 try:
     import util
     import config
     import userprofile
     import dirmonitor
+    import mozilla_bookmarks
 except:
     from sabayon import util
     from sabayon import config
     from sabayon import userprofile
     from sabayon import dirmonitor
+    from sabayon import mozilla_bookmarks
 
 class file_state:
     (UNKNOWN,
@@ -45,10 +46,15 @@ class file_state:
 
 
 firefox_rel_path = ".mozilla/firefox"
+
 profiles_ini_rel_path = os.path.join(firefox_rel_path, "profiles.ini")
+
 sabayon_pref_rel_path = os.path.join(firefox_rel_path, "sabayon-prefs.js")
 sabayon_mandatory_pref_rel_path = os.path.join(firefox_rel_path, "sabayon-mandatory-prefs.js")
 
+sabayon_bookmark_rel_path = os.path.join(firefox_rel_path, "sabayon-bookmarks.html")
+sabayon_mandatory_bookmark_rel_path = os.path.join(firefox_rel_path, "sabayon-mandatory-bookmarks.html")
+
 LOG_OPERATION           = 0x00001
 LOG_CHANGE              = 0x00002
 LOG_IGNORED_CHANGE      = 0x00004
@@ -66,6 +72,21 @@ def dprint(mask, fmt, *args):
 def dwarn(fmt, *args):
     print >> sys.stderr, "WARNING " + fmt % args
 
+bookmark_exclude_attrs = [re.compile("^id$"),
+                          re.compile("^last_"),
+                          re.compile("^add_"),
+                          #re.compile("^icon$"),
+                          ]
+
+def filter_attr(attr, exclude):
+    if not exclude:
+        return False
+
+    for regexp in exclude:
+        if regexp.search(attr):
+            return True
+    return False
+
 class MozillaChange(userprofile.ProfileChange):
     (
         CREATED,
@@ -119,12 +140,14 @@ gobject.type_register(MozillaChange)
 class MozillaDelegate(userprofile.SourceDelegate):
     def __init__ (self, source):
         dprint(LOG_OPERATION, "Delegate construction")
-        userprofile.SourceDelegate.__init__ (self, _("Firefox"), source, ".mozilla")
+        userprofile.SourceDelegate.__init__ (self, "Firefox", source, ".mozilla")
         self.source = source
         self.delegate = self
         self.home_dir = util.get_home_dir()
         self.committed_prefs = {}
         self.committed_mandatory_prefs = {}
+        self.committed_bookmarks = mozilla_bookmarks.BookmarkFolder("Null", None)
+        self.committed_mandatory_bookmarks = mozilla_bookmarks.BookmarkFolder("Null", None)
         self.ini_file = None
 
     def get_full_path(self, path):
@@ -133,6 +156,26 @@ class MozillaDelegate(userprofile.SourceDelegate):
     def get_profiles_ini_path(self):
         return self.get_full_path(profiles_ini_rel_path)
 
+    def get_path_description (self, path):
+        type = get_type_from_path(path)
+
+        if   type == FirefoxProfileFile.TYPE_PREFS:
+            return _("Web browser preferences")
+        elif type == FirefoxProfileFile.TYPE_BOOKMARK:
+            return _("Web browser bookmarks")
+        elif type == FirefoxProfileFile.TYPE_PROFILE_INI:
+            return _("Web browser profile list")
+        else:
+            # XXX
+            basename = os.path.basename(path)
+    
+            if   basename.endswith("prefs.js"):
+                return _("Web browser preferences")
+            elif basename.endswith("bookmarks.html"):
+                return _("Web browser bookmarks")
+            else:
+                return path
+
     def load_profiles_ini(self):
         if not self.ini_file:
             self.ini_file = FirefoxProfilesIni(self.home_dir, profiles_ini_rel_path)
@@ -177,20 +220,31 @@ class MozillaDelegate(userprofile.SourceDelegate):
             dprint(LOG_CHANGE, "%s profile file: %s",
                    dirmonitor.event_to_string(change.event), rel_path)
 
-            # XXX - jrd, we shouldn't have to create a file just to get its type
             profile_file = profile.add_file(rel_path)
             profile_file_type = profile_file.get_type()
 
             if profile_file_type == FirefoxProfileFile.TYPE_PREFS:
+                assert isinstance(profile_file, JavascriptPrefsFile)
                 if change.event == dirmonitor.CREATED or \
                    change.event == dirmonitor.CHANGED:
-                    if not isinstance(profile_file, JavascriptPrefsFile):
-                        profile_file = profile.add_file(rel_path,JavascriptPrefsFile(self.home_dir, rel_path))
                     cat_file(profile_file.get_full_path())
                     profile_file.read()
                     profile_file.emit_changes(self.source, self.delegate)
                     return True
                 elif change.event == dirmonitor.DELETED:
+                    # XXX - what to do here?
+                    pass
+                else:
+                    raise ValueError
+            elif profile_file_type == FirefoxProfileFile.TYPE_BOOKMARK:
+                assert isinstance(profile_file, BookmarksFile)
+                if change.event == dirmonitor.CREATED or \
+                   change.event == dirmonitor.CHANGED:
+                    profile_file.read()
+                    profile_file.emit_changes(self.source, self.delegate)
+                    return True
+                elif change.event == dirmonitor.DELETED:
+                    # XXX - what to do here?
                     pass
                 else:
                     raise ValueError
@@ -212,17 +266,27 @@ class MozillaDelegate(userprofile.SourceDelegate):
         return True
 
     def commit_change(self, change, mandatory = False):
-        dprint(LOG_CHANGE, "Commiting (mandatory = %s) key = %s value = %s event = %s",
-                mandatory, change.key, change.value, change.event)
+        if isinstance(change, MozillaChange):
+            dprint(LOG_CHANGE, "Commiting preference (mandatory = %s) key = %s value = %s event = %s",
+                   mandatory, change.key, change.value, change.event)
 
-        # XXX - jrd we need to mark here what gets written in sync_changes
-        assert isinstance(change, MozillaChange)
-        if mandatory:
-            self.committed_mandatory_prefs[change.get_key()] = \
-                JavascriptPreference(change.get_type(), change.get_key(), change.get_value())
-        else:
-            self.committed_prefs[change.get_key()] = \
-                JavascriptPreference(change.get_type(), change.get_key(), change.get_value())
+            if mandatory:
+                self.committed_mandatory_prefs[change.get_key()] = \
+                    JavascriptPreference(change.get_type(), change.get_key(), change.get_value())
+            else:
+                self.committed_prefs[change.get_key()] = \
+                    JavascriptPreference(change.get_type(), change.get_key(), change.get_value())
+        elif isinstance(change, BookmarkChange):
+            bookmark_path = change.get_bookmark_path()
+            entry = change.get_entry()
+
+            dprint(LOG_CHANGE, "Commiting bookmark (mandatory = %s) path = %s url = %s event = %s",
+                   mandatory, entry.path_as_string(), entry.get_url(), change.event)
+
+            if mandatory:
+                self.committed_mandatory_bookmarks.add_path_entry(entry.path(), entry)
+            else:
+                self.committed_bookmarks.add_path_entry(entry.path(), entry)
                 
 
     def start_monitoring (self):
@@ -269,7 +333,31 @@ class MozillaDelegate(userprofile.SourceDelegate):
                                          {"file_type" : FirefoxProfileFile.TYPE_PREFS,
                                           "mandatory"    : True})
 
-    def apply(self):
+            if len(self.committed_bookmarks.entries) > 0:
+                bookmark_rel_path = sabayon_bookmark_rel_path
+                dprint(LOG_SYNC, "sync_changes: storing committed_bookmarks to %s", bookmark_rel_path)
+                bookmark_file = BookmarksFile(self.home_dir, bookmark_rel_path)
+                bookmark_file.set_root(self.committed_bookmarks)
+                bookmark_file.write(exclude_attrs=bookmark_exclude_attrs)
+                self.source.storage.add(bookmark_rel_path, self.home_dir, self.name,
+                                         {"file_type" : FirefoxProfileFile.TYPE_BOOKMARK,
+                                          "mandatory"    : False})
+
+            if len(self.committed_mandatory_bookmarks.entries) > 0:
+                bookmark_rel_path = sabayon_mandatory_bookmark_rel_path
+                dprint(LOG_SYNC, "sync_changes: storing mandatory committed_bookmarks to %s", bookmark_rel_path)
+                bookmark_file = BookmarksFile(self.home_dir, bookmark_rel_path)
+                bookmark_file.set_root(self.committed_mandatory_bookmarks)
+                bookmark_file.write(exclude_attrs=bookmark_exclude_attrs)
+                self.source.storage.add(bookmark_rel_path, self.home_dir, self.name,
+                                         {"file_type" : FirefoxProfileFile.TYPE_BOOKMARK,
+                                          "mandatory"    : True})
+
+    def set_enforce_mandatory (self, enforce):
+        # Nothing to do here
+        pass
+
+    def apply(self, is_sabayon_session):
         ini_files = []
         pref_files = []
         bookmark_files = []
@@ -312,27 +400,51 @@ class MozillaDelegate(userprofile.SourceDelegate):
             else:
                 dprint(LOG_APPLY, "apply: but there isn't an ini file in the profile!")
 
+        # --------------------
         if sabayon_pref_rel_path in pref_files:
             dprint(LOG_APPLY, "extracting %s" % sabayon_pref_rel_path)
             self.source.storage.extract(sabayon_pref_rel_path, self.home_dir, True)
-            apply_pref = JavascriptPrefsFile(self.home_dir, path)
+            apply_pref = JavascriptPrefsFile(self.home_dir, sabayon_pref_rel_path)
             apply_pref.read()
         else:
             apply_pref = None
 
         if sabayon_mandatory_pref_rel_path in pref_files:
-            dprint(LOG_APPLY, ">>> extracting %s" % sabayon_mandatory_pref_rel_path)
+            dprint(LOG_APPLY, "extracting %s" % sabayon_mandatory_pref_rel_path)
             self.source.storage.extract(sabayon_mandatory_pref_rel_path, self.home_dir, True)
-            mandatory_apply_pref = JavascriptPrefsFile(self.home_dir, path)
+            mandatory_apply_pref = JavascriptPrefsFile(self.home_dir, sabayon_mandatory_pref_rel_path)
             mandatory_apply_pref.read()
         else:
             mandatory_apply_pref = None
 
+        # --------------------
+
+        if sabayon_bookmark_rel_path in bookmark_files:
+            dprint(LOG_APPLY, "extracting %s" % sabayon_bookmark_rel_path)
+            self.source.storage.extract(sabayon_bookmark_rel_path, self.home_dir, True)
+            apply_bookmark = BookmarksFile(self.home_dir, sabayon_bookmark_rel_path)
+            apply_bookmark.read()
+        else:
+            apply_bookmark = None
+
+        if sabayon_mandatory_bookmark_rel_path in bookmark_files:
+            dprint(LOG_APPLY, "extracting %s" % sabayon_mandatory_bookmark_rel_path)
+            self.source.storage.extract(sabayon_mandatory_bookmark_rel_path, self.home_dir, True)
+            mandatory_apply_bookmark = BookmarksFile(self.home_dir, sabayon_mandatory_bookmark_rel_path)
+            mandatory_apply_bookmark.read()
+        else:
+            mandatory_apply_bookmark = None
+
+
+        # --------------------
+
         # Now merge the javascript pref files
         # XXX - iterate over all target profiles
         for profile in self.ini_file.get_profiles():
             dprint(LOG_APPLY, "apply: applying to profile %s", profile.attrs)
             profile_rel_dir = profile.get_rel_dir()
+
+            # --------------------
             target_pref_rel_path = os.path.join(profile_rel_dir, "prefs.js")
             target_pref = JavascriptPrefsFile(self.home_dir, target_pref_rel_path)
             target_pref.read()
@@ -352,10 +464,25 @@ class MozillaDelegate(userprofile.SourceDelegate):
             if apply_pref or mandatory_apply_pref:
                 target_pref.write()
 
-        # Now merge the bookmarks
-        for path in bookmark_files:
-            # XXX - merge not implemented.
-            pass
+            # --------------------
+            target_bookmark_rel_path = os.path.join(profile_rel_dir, "bookmarks.html")
+            target_bookmark = BookmarksFile(self.home_dir, target_bookmark_rel_path)
+            target_bookmark.read()
+
+            if apply_bookmark:
+                mandatory = False
+                dprint(LOG_APPLY, "apply: applying src profile %s to target profile %s, mandatory=%s",
+                       sabayon_bookmark_rel_path, target_bookmark_rel_path, mandatory)
+                target_bookmark.merge(apply_bookmark, mandatory)
+
+            if mandatory_apply_bookmark:
+                mandatory = True
+                dprint(LOG_APPLY, "apply: applying src profile %s to target profile %s, mandatory=%s",
+                       sabayon_bookmark_rel_path, target_bookmark_rel_path, mandatory)
+                target_bookmark.merge(mandatory_apply_bookmark, mandatory)
+
+            if apply_bookmark or mandatory_apply_bookmark:
+                target_bookmark.write()
 
         # Finally extract any other file
         for path in other_files:
@@ -397,128 +524,6 @@ class BadIniFileError(Exception):
         return self.problem
     
 
-# ------ Class DictCompare ------
-
-class DictCompare:
-    def __init__(self, a, b):
-        self.a = a
-        self.b = b
-
-    def compare(self):
-        ''' Given two dictionaries a,b analyze them for their
-        differences and similarities.
-
-        intersection - keys shared between a and b
-        only_a       - keys only present in a
-        only_b       - keys only present in b
-        equal        - keys present in both a and b whole values are equal
-        not_equal    - keys present in both a and b whole values are not equal'''
-
-        self.keys_a = self.a.keys()
-        self.keys_b = self.b.keys()
-        
-        self.intersection = []
-        self.only_a = []
-        self.only_b = []
-        self.equal = []
-        self.not_equal = []
-        self._add = {}
-        self._del = {}
-        self._mod = {}
-
-        for k in self.keys_a:
-            if self.b.has_key(k):
-                self.intersection.append(k)
-            else:
-                self.only_a.append(k)
-
-        for k in self.keys_b:
-            if not self.a.has_key(k):
-                self.only_b.append(k)
-                
-        for k in self.intersection:
-            if self.a[k] == self.b[k]:
-                self.equal.append(k)
-            else:
-                self.not_equal.append(k)
-                
-    def intersection(self):
-        'return list of keys shared between a and b'
-        return self.intersection
-
-    def only_a(self):
-        'return list of keys only present in a'
-        return self.only_a
-
-    def only_b(self):
-        'return list of keys only present in b'
-        return self.only_b
-
-    def equal(self):
-        'return list of keys present in both a and b whole values are equal'
-        return self.equal
-
-    def not_equal(self):
-        'return list of keys present in both a and b whole values are not equal'
-        return self.not_equal
-
-    def get_change_set(self, dict_lhs, dict_rhs):
-        '''Return changes necessary to make dict_lhs equivalent to dict_rhs,
-        (e.g. lhs = rhs), the two dictionary parameters are specified as
-        either the string 'a' or the string 'b' corresponding to the parameters
-        this class was created with.
-        
-        Return value is a dictionary with 3 keys (add, del, mod) whose values
-        are dictionaries containing containing (key,value) pairs to add,
-        delete, or modify respectively in dict_lhs.'''
-
-        if dict_lhs == dict_rhs or dict_lhs not in "ab" or dict_rhs not in "ab":
-            raise ValueError
-
-        if dict_lhs == 'a':
-            a = self.a
-            b = self.b
-            only_a = self.only_a
-            only_b = self.only_b
-        elif dict_lhs == 'b':
-            a = self.b
-            b = self.a
-            only_a = self.only_b
-            only_b = self.only_a
-        else:
-            raise ValueError
-
-        self._add = {}
-        for k in only_b:
-            self._add[k] = b[k]
-
-        self._del = {}
-        for k in only_a:
-            self._del[k] = a[k]
-
-        self._mod = {}
-        for k in self.not_equal:
-            self._mod[k] = b[k]
-
-        change_set = {'add':self._add, 'del':self._del, 'mod':self._mod}
-        return change_set
-
-
-    def is_equal(self):
-        if len(self.only_a) == 0 and len(self.only_b) == 0 and len(self.not_equal) == 0:
-            return True
-        else:
-            return False
-
-    def dump(self):
-        'Print the results of the dictionary comparision'
-        dprint(LOG_DATA, "intersection = %s" % ",".join(self.intersection))
-        dprint(LOG_DATA, "only a = %s" % ",".join(self.only_a))
-        dprint(LOG_DATA, "only b = %s" % ",".join(self.only_b))
-        dprint(LOG_DATA, "equal = %s" % ",".join(self.equal))
-        dprint(LOG_DATA, "not equal = %s" % ",".join(self.not_equal))
-
-
 # ------ Class FirefoxProfileFile ------
 
 class FirefoxProfileFile:
@@ -528,11 +533,18 @@ class FirefoxProfileFile:
      TYPE_BOOKMARK
      ) = range(4)
 
-    def __init__(self, rel_path):
+    def __init__(self, home_dir, rel_path):
+        self.home_dir = home_dir
         self.rel_path = rel_path
         self.attrs = {}
         self.file_type = get_type_from_path(rel_path)
 
+    def get_full_path(self):
+        return os.path.join(self.home_dir, self.rel_path)
+
+    def get_rel_path(self):
+        return self.rel_path
+
     def type_to_string(self, type):
         if type == FirefoxProfileFile.TYPE_UNKNOWN:
             return "UNKNOWN"
@@ -577,16 +589,11 @@ class JavascriptPreference:
 class JavascriptPrefsFile(FirefoxProfileFile):
     def __init__(self, home_dir, rel_path):
         dprint(LOG_OPERATION, "JavascriptPrefsFile: created (%s)", rel_path)
-        FirefoxProfileFile.__init__(self, rel_path)
+        FirefoxProfileFile.__init__(self, home_dir, rel_path)
         self.file_state = file_state.UNKNOWN
-        self.home_dir = home_dir
         self.prefs = {}
         self.prev_prefs = {}
 
-
-    def get_full_path(self):
-        return os.path.join(self.home_dir, self.rel_path)
-
     def get_file_state(self):
         return self.file_state
 
@@ -628,13 +635,15 @@ class JavascriptPrefsFile(FirefoxProfileFile):
                                                            src_pref.get_value())
                 
 
-    def read(self):
+    def read(self, full_path = None):
         self.prev_prefs = self.get_prefs()
 
+        if not full_path:
+            full_path = self.get_full_path()
         dprint(LOG_OPERATION, "read profile prefs (%s)", self.get_full_path())
         self.file_state = file_state.UNKNOWN
         try:
-            fd = open(self.get_full_path())
+            fd = open(full_path)
         except IOError, e:
             if e.errno == errno.ENOENT:
                 self.file_state = file_state.NOT_FOUND
@@ -654,10 +663,9 @@ class JavascriptPrefsFile(FirefoxProfileFile):
     def emit_changes(self, source, delegate):
         cur_prefs = self.get_prefs()
 
-        dc = DictCompare(self.prev_prefs, cur_prefs)
+        dc = util.DictCompare(self.prev_prefs, cur_prefs)
         dc.compare()
         cs = dc.get_change_set('a', 'b')
-        dump_change_set(cs)
 
         _add = cs['add']
         _del = cs['del']
@@ -673,17 +681,24 @@ class JavascriptPrefsFile(FirefoxProfileFile):
         emit_changes(_del.items(), MozillaChange.DELETED)
         emit_changes(_mod.items(), MozillaChange.CHANGED)
 
-        self.prev_prefs = cur_prefs
-
+        
 
     def kill_comments(self):
-        slash_comment_re = re.compile("//.*$", re.MULTILINE)
-        hash_comment_re = re.compile("#.*$", re.MULTILINE)
-        c_comment_re = re.compile("/\*.*?\*/", re.MULTILINE | re.DOTALL)
+        def not_in_string (regexp):
+            double = '\\"(?:\\\\.|[^\"\\\\])*\\"'
+            single = "\\'(?:\\\\.|[^\"\\\\])*\\'"
+            return "(" + double + "|" + single + ")(?:" + regexp + ")"
+        def match (matchobj):
+            return matchobj.group(1)
+        
+            
+        slash_comment_re = re.compile(not_in_string ("//.*$"), re.MULTILINE)
+        hash_comment_re = re.compile(not_in_string ("#.*$"), re.MULTILINE)
+        c_comment_re = re.compile(not_in_string ("/\*.*?\*/"), re.MULTILINE | re.DOTALL)
 
-        self.filebuf = slash_comment_re.sub("", self.filebuf)
-        self.filebuf = hash_comment_re.sub("", self.filebuf)
-        self.filebuf = c_comment_re.sub("", self.filebuf)
+        self.filebuf = slash_comment_re.sub(match, self.filebuf)
+        self.filebuf = hash_comment_re.sub(match, self.filebuf)
+        self.filebuf = c_comment_re.sub(match, self.filebuf)
 
     def parse(self):
         start = 0;
@@ -718,8 +733,9 @@ class JavascriptPrefsFile(FirefoxProfileFile):
 # ------ Class FirefoxProfile ------
 
 class FirefoxProfile:
-    def __init__(self, section, rel_dir):
+    def __init__(self, section, home_dir, rel_dir):
         self.section = section
+        self.home_dir = home_dir
         self.rel_dir = rel_dir
         self.attrs = {}
         self.files = {}
@@ -742,15 +758,22 @@ class FirefoxProfile:
     def get_dir(self):
         return self.get_attr("path")
 
-    def add_file(self, rel_path, object=None):
-        # XXX - jrd, passing object in is ugly & awkward, find more elegant solution
-        if not object:
-            object = FirefoxProfileFile(rel_path)
-            file = self.files.setdefault(rel_path, object)
+    def add_file(self, rel_path):
+        object = self.files.get(rel_path, None)
+        if object:
+            return(object)
+
+
+        file_type = get_type_from_path(rel_path)
+
+        if file_type == FirefoxProfileFile.TYPE_PREFS:
+            object = JavascriptPrefsFile(self.home_dir, rel_path)
+        elif file_type == FirefoxProfileFile.TYPE_BOOKMARK:
+            object = BookmarksFile(self.home_dir, rel_path)
         else:
-            file = object
-            self.files[rel_path] = object
-        return file
+            object = FirefoxProfileFile(self.home_dir, rel_path)
+        self.files[rel_path] = object
+        return object
 
     def del_file(self, rel_path):
         if rel_path in self.files:
@@ -792,11 +815,18 @@ class FirefoxProfilesIni:
         dprint(LOG_OPERATION, "FirefoxProfilesIni.load_profiles()")
         for profile in self.get_profiles():
             profile_rel_dir = profile.get_rel_dir()
+
             pref_rel_path = os.path.join(profile_rel_dir, "prefs.js")
             dprint(LOG_OPERATION, "FirefoxProfilesIni.load_profiles() pref=%s", pref_rel_path)
-            pref = JavascriptPrefsFile(self.home_dir, pref_rel_path)
-            profile.add_file(pref_rel_path, pref)
-            pref.read()
+            profile_file = profile.add_file(pref_rel_path)
+            profile_file.read()
+
+            bookmark_rel_path = os.path.join(profile_rel_dir, "bookmarks.html")
+            dprint(LOG_OPERATION, "FirefoxProfilesIni.load_profiles() bookmark=%s", bookmark_rel_path)
+            profile_file = profile.add_file(bookmark_rel_path)
+            profile_file.read()
+
+
 
     def read(self):
         dprint(LOG_OPERATION, "FirefoxProfilesIni.read() path = %s",
@@ -836,9 +866,9 @@ class FirefoxProfilesIni:
                 
                 name = self.ini.get(section, "Name")
                 if name in self.profiles:
-                    raise BadIniFileError(_("duplicate name(%s) in section %s") %
+                    raise BadIniFileError(_("duplicate name(%(name)s) in section %(section)s") %
                                           (name, section))
-                profile = FirefoxProfile(section, self.rel_dir)
+                profile = FirefoxProfile(section, self.home_dir, self.rel_dir)
                 self.profiles[name] = profile
                 for (key, value) in self.ini.items(section):
                     profile.set_attr(key, value)
@@ -872,6 +902,205 @@ class FirefoxProfilesIni:
         else:
             return self.profiles.values()
 
+# ------ Class BookmarkChange ------
+
+class BookmarkChange(userprofile.ProfileChange):
+    (
+        CREATED,
+        DELETED,
+        CHANGED
+    ) = range(3)
+    
+    def __init__ (self, source, delegate, entry, event):
+        userprofile.ProfileChange.__init__ (self, source, delegate)
+        
+        assert event == self.CREATED or \
+               event == self.DELETED or \
+               event == self.CHANGED
+        
+        self.entry = entry
+        self.event = event
+        self.attrs = {}
+
+    def set_attr(self, attr, value):
+        self.attrs[attr] = value
+
+    def get_attr(self, attr):
+        return self.attrs[attr]
+
+    def get_bookmark_path(self):
+        return self.entry.path_as_string()
+
+    def get_entry(self):
+        return self.entry
+
+    def get_url(self):
+        return self.entry.get_url()
+
+    def get_id(self):
+        return self.get_bookmark_path()
+
+    def get_short_description(self):
+        url = self.get_url()
+        bookmark_path = self.get_bookmark_path()
+        
+        if self.event == self.CREATED:
+            # XXX - don't test for url, use type
+            if url:
+                return _("Mozilla bookmark created '%s' -> '%s'") % (bookmark_path, url)
+            else:
+                return _("Mozilla bookmark folder created '%s'") % (bookmark_path)
+        elif self.event == self.DELETED:
+            if url:
+                return _("Mozilla bookmark deleted '%s'") % (bookmark_path)
+            else:
+                return _("Mozilla bookmark folder deleted '%s'") % (bookmark_path)
+        elif self.event == self.CHANGED:
+            if url:
+                return _("Mozilla bookmark changed '%s' '%s'") % (bookmark_path, url)
+            else:
+                return _("Mozilla bookmark folder changed '%s'") % (bookmark_path)
+                
+        else:
+            raise ValueError
+
+gobject.type_register(BookmarkChange)
+
+# ------ Class BookmarksFile ------
+
+class BookmarksFile(FirefoxProfileFile):
+    def __init__(self, home_dir, rel_path):
+        dprint(LOG_OPERATION, "BookmarksFile: created (%s)", rel_path)
+        FirefoxProfileFile.__init__(self, home_dir, rel_path)
+        self.parser = mozilla_bookmarks.BookmarkHTMLParser()
+        self.root = mozilla_bookmarks.BookmarkFolder("Null", None)
+        self.parser.set_root(self.root)
+        self.prev_root = self.parser.get_root()
+        self.file_state = file_state.UNKNOWN
+
+    def get_full_path(self):
+        return os.path.join(self.home_dir, self.rel_path)
+
+    def get_rel_path(self):
+        return self.rel_path
+
+    def get_root(self):
+        return self.root
+
+    def set_root(self, root):
+        self.root = root
+        self.parser.set_root(self.root)
+
+    def read(self):
+        self.prev_root = self.parser.get_root()
+        self.root = mozilla_bookmarks.BookmarkFolder("Null", None)
+        self.parser.set_root(self.root)
+        self.file_state = file_state.UNKNOWN
+        full_path = self.get_full_path()
+        dprint(LOG_OPERATION, "BookmarksFile: read (%s)", full_path)
+        try:
+            fd = open(full_path)
+        except IOError, e:
+            if e.errno == errno.ENOENT:
+                self.file_state = file_state.NOT_FOUND
+                return
+            else:
+                self.file_state = file_state.SYS_ERROR
+                raise
+
+        self.file_state = file_state.VALID
+        self.parser.feed(fd.read())
+        self.parser.close()
+        self.root = self.parser.get_root()
+        
+    def write(self, full_path=None, exclude_attrs=None):
+        if not full_path:
+            full_path = self.get_full_path()
+        dir = os.path.dirname(full_path)
+        if not os.path.exists(dir):
+            os.makedirs(dir)
+        dprint(LOG_OPERATION, "MozillaBookmark: writing file (%s)", full_path)
+        fd = open(full_path, "w")
+        fd.write('''
+<!DOCTYPE NETSCAPE-Bookmark-file-1>
+<!-- This is an automatically generated file. (Created by %s, version %s)
+     It will be read and overwritten.
+     DO NOT EDIT! -->
+<META HTTP-EQUIV="Content-Type" CONTENT="text/html; charset=UTF-8">
+<TITLE>Bookmarks</TITLE>
+<H1 LAST_MODIFIED="%.0f">Bookmarks</H1>
+
+<DL><p>
+''' % (config.PACKAGE, config.VERSION, time.time()))
+
+        def visit(entry, type, path, data):
+            indent = "    "
+            level = len(path)
+
+            if type == mozilla_bookmarks.TYPE_FOLDER:
+                fd.write("%s<DT><H3" % (indent*level))
+                for attr, value in entry.attrs.items():
+                    if not filter_attr(attr, exclude_attrs):
+                        fd.write(" %s=\"%s\"" % (attr, value))
+                fd.write(">%s</H3>\n" % (entry.name))
+                fd.write("%s<DL><p>\n" % (indent*level))
+            elif type == mozilla_bookmarks.TYPE_BOOKMARK:
+                fd.write("%s<DT><A" % (indent*level))
+                for attr, value in entry.attrs.items():
+                    if not filter_attr(attr, exclude_attrs):
+                        fd.write(" %s=\"%s\"" % (attr, value))
+                fd.write(">%s</A>\n" % (entry.name))
+            elif type == mozilla_bookmarks.TYPE_FOLDER_END:
+                fd.write("%s</DL><p>\n" % (indent*level))
+            else:
+                raise ValueError
+
+        self.root.traverse(visit)
+        fd.close()
+        
+    def convert_to_dict(self, root):
+        result = {}
+
+        def visit(entry, type, path, data):
+            bookmark_path = entry.path_as_string()
+            result[bookmark_path] = entry
+
+        root.traverse(visit)
+        return result
+    
+    def emit_changes(self, source, delegate):
+        prev_dict = self.convert_to_dict(self.prev_root)
+        cur_dict  = self.convert_to_dict(self.root)
+
+        dc = util.DictCompare(prev_dict, cur_dict)
+        dc.compare()
+        cs = dc.get_change_set('a', 'b')
+
+        _add = cs['add']
+        _del = cs['del']
+        _mod = cs['mod']
+
+        def emit_changes(items, event):
+            for bookmark_path, entry in items:
+                source.emit_change(
+                    BookmarkChange(source, delegate, entry, event))
+
+        emit_changes(_add.items(), BookmarkChange.CREATED)
+        emit_changes(_del.items(), BookmarkChange.DELETED)
+        emit_changes(_mod.items(), BookmarkChange.CHANGED)
+
+
+    def merge(self, src, mandatory):
+        def visit(entry, type, path, data):
+            if type == mozilla_bookmarks.TYPE_FOLDER_END:
+                return
+            dst_entry = self.root.lookup_path(path)
+            if not dst_entry or mandatory:
+                self.root.add_path_entry(path, entry)
+
+        src.root.traverse(visit)
+
+
 # ------ Utility Functions ------
 
 def get_type_from_path(rel_path):
@@ -881,76 +1110,11 @@ def get_type_from_path(rel_path):
         return FirefoxProfileFile.TYPE_PREFS
     elif basename == "bookmarks.html":
         return FirefoxProfileFile.TYPE_BOOKMARK
+    elif basename == "profiles.ini":
+        return FirefoxProfileFile.TYPE_PROFILE_INI
     else:
         return FirefoxProfileFile.TYPE_UNKNOWN
 
-
-# XXX - this needs more logic to distinquish mozilla, firefox, versions, etc.
-# basically its just hardcoded at the moment.
-def GetProfileIniFile():
-    filename = os.path.expanduser("~/%s" % profiles_ini_rel_path)
-
-    # XXX - should caller thow error instead?
-    if not os.path.exists(filename):
-        raise FileNotFoundError(filename)
-
-    return(filename)
-
-def insert_prefs_into_file(path, prefs):
-    (wfd, tmppath) = tempfile.mkstemp(dir=os.path.dirname(path))
-
-    for line in open(path):
-        wfd.write(line)
-
-    for key, value in prefs.items():
-        wfd.write("user_pref(\"%s\", %s);\n" % (key, value))
-
-    wfd.close()
-    os.rename(tmppath, path)
-
-# XXX - this does not deal with comments
-def remove_prefs_from_file(path, prefs):
-    (wfd, tmppath) = tempfile.mkstemp(dir=os.path.dirname(path))
-
-    for line in open(path):
-        match = pref_re.search(line)
-        if match:
-            key = match.group(2)
-            if type(prefs) is types.DictType:
-                if not prefs.has_key(key):
-                    wfd.write(line)
-            elif type(prefs) is types.ListType:
-                if not key in prefs:
-                    wfd.write(line)
-            else:
-                raise ValueError
-        else:
-            wfd.write(line)
-
-    wfd.close()
-    os.rename(tmppath, path)
-
-
-def dump_change_set(cs):
-    _add = cs['add']
-    _del = cs['del']
-    _mod = cs['mod']
-
-    if len(_add.keys()):
-        dprint(LOG_DATA, "Key/Values to ADD")
-        for k in _add.keys():
-            dprint(LOG_DATA, "    %s=%s" % (k, _add[k]))
-
-    if len(_del.keys()):
-        dprint(LOG_DATA, "Keys to DELETE")
-        for k in _del.keys():
-            dprint(LOG_DATA, "    %s=%s" % (k, _del[k]))
-
-    if len(_mod.keys()):
-        dprint(LOG_DATA, "Key/Values to Modify")
-        for k in _mod.keys():
-            dprint(LOG_DATA, "    %s=%s" % (k, _mod[k]))
-
 def cat_file(path):
     if os.path.isfile(path):
         dprint(LOG_FILE_CONTENTS, "==== %s ====" % path)
@@ -970,66 +1134,3 @@ def run_unit_tests():
     test_prefs = {'foo':'"bar"', 'uno':'1'}
 
     dprint(LOG_OPERATION, "In mozillaprofile tests")
-
-    try:
-        profile_ini_file = GetProfileIniFile()
-    except FileNotFoundError, e:
-        print _("No such profile ini file: %s") % e.filename
-        return
-
-    dprint(LOG_OPERATION, "ini file = %s" % profile_ini_file)
-
-    profiles_ini = FirefoxProfilesIni(os.path.expanduser("~"), profile_ini_file)
-    profiles_ini.read()
-    profiles_ini.parse_sections()
-    default_path = profiles_ini.get_default_path()
-    dprint(LOG_OPERATION, "default_path = %s" % default_path)
-
-    prefs_path = "%s/%s" % (default_path, "prefs.js")
-
-    # make sure we're working with a clean file copy
-    remove_prefs_from_file(prefs_path, test_prefs)
-
-    pref = JavascriptPrefsFile(prefs_path)
-    pref.open()
-    pref.kill_comments()
-    pref.parse()
-    prev_prefs = pref.get_prefs()
-    
-    insert_prefs_into_file(prefs_path, test_prefs)
-
-    pref = JavascriptPrefsFile(prefs_path)
-    pref.open()
-    pref.kill_comments()
-    pref.parse()
-    cur_prefs = pref.get_prefs()
-
-    dc = DictCompare(prev_prefs, cur_prefs)
-    dc.compare()
-
-    cs = dc.get_change_set('a', 'b')
-    dprint(LOG_DATA, "a <-- b")
-    dump_change_set(cs)
-
-    dc = DictCompare(test_prefs, cs['add'])
-    dc.compare()
-    assert dc.is_equal() == True
-
-    test_prefs['newkey'] = 'new'
-    dc = DictCompare(test_prefs, cs['add'])
-    dc.compare()
-    assert dc.is_equal() == False
-
-    del test_prefs['newkey']
-    dc = DictCompare(test_prefs, cs['add'])
-    dc.compare()
-    assert dc.is_equal() == True
-
-    test_prefs['uno'] = '2'
-    dc = DictCompare(test_prefs, cs['add'])
-    dc.compare()
-    assert dc.is_equal() == False
-
-    remove_prefs_from_file(prefs_path, test_prefs)
-
-
