@@ -1,5 +1,3 @@
-#!/usr/bin/env python
-
 #
 # Copyright (C) 2005 Red Hat, Inc.
 #
@@ -18,6 +16,7 @@
 # Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 #
 
+import os
 import gconf
 import gconfsource
 
@@ -33,12 +32,16 @@ except:
 def dprint (fmt, *args):
     util.debug_print (util.DEBUG_PANELDELEGATE, fmt % args)
 
+PANEL_LAUNCHER_DIR = ".gnome2/panel2.d/default/launchers"
+
 class PanelChange (userprofile.ProfileChange):
     def __init__ (self, source, delegate, id):
         userprofile.ProfileChange.__init__ (self, source, delegate)
         self.id = id
     def get_id (self):
         return self.id
+    def commit_change (self, mandatory):
+        pass
 
 #
 # FIXME: these short descriptions are lame, should be:
@@ -75,12 +78,23 @@ class PanelObjectAddedChange (PanelChange):
         PanelChange.__init__ (self, source, delegate, id)
     def get_short_description (self):
         return _("Panel object '%s' added") % self.id
+    def commit_change (self, mandatory):
+        # Might have to commit a launcher file
+        launcher = self.delegate.client.get_string (PANEL_KEY_BASE + "/objects/" + self.id + "/launcher_location")
+        if launcher and launcher[0] != '/':
+            file = PANEL_LAUNCHER_DIR + "/" + launcher
+            self.source.storage.add (file, self.delegate.home_dir, self.delegate.name)
     
 class PanelObjectRemovedChange (PanelChange):
     def __init__ (self, source, delegate, id):
         PanelChange.__init__ (self, source, delegate, id)
     def get_short_description (self):
         return _("Panel object '%s' removed") % self.id
+    def commit_change (self, mandatory):
+        launcher = self.delegate.client.get_string (PANEL_KEY_BASE + "/objects/" + self.id + "/launcher_location")
+        if launcher and launcher[0] != '/':
+            file = PANEL_LAUNCHER_DIR + "/" + launcher
+            self.source.storage.remove (file)
 
 class PanelDelegate (userprofile.SourceDelegate):
     class PanelThing:
@@ -97,7 +111,7 @@ class PanelDelegate (userprofile.SourceDelegate):
             # self.name        = self.client.get_string (PANEL_KEY_BASE + "/toplevels/" + toplevel_id + "/name")
             # self.orientation = self.client.get_string (PANEL_KEY_BASE + "/toplevels/" + toplevel_id + "/orientation")
             # self.expand      = self.client.get_bool   (PANEL_KEY_BASE + "/toplevels/" + toplevel_id + "/expand")
-        
+
     class PanelApplet (PanelThing):
         def __init__ (self, id, added = False, removed = False):
             PanelDelegate.PanelThing.__init__ (self, id, added, removed)
@@ -118,21 +132,31 @@ class PanelDelegate (userprofile.SourceDelegate):
         userprofile.SourceDelegate.__init__ (self, _("Panel"), source, PANEL_KEY_BASE)
         self.client = gconf.client_get_default ()
 
+        self.home_dir = util.get_home_dir()
         self.toplevels = {}
         self.applets = {}
         self.objects = {}
-        self.__read_panel_config ()
 
     def __read_panel_config (self):
+        dprint ("Reading initial panel config");
+        
+        dprint ("Toplevels:");
         for id in self.client.get_list (PANEL_KEY_BASE + "/general/toplevel_id_list", gconf.VALUE_STRING):
             if not self.toplevels.has_key (id):
-                self.toplevels[id] = PanelDelegate.PanelToplevel (id, False)
+                dprint ("  %s", id);
+                self.toplevels[id] = PanelDelegate.PanelToplevel (id)
+                
+        dprint ("Applets:");
         for id in self.client.get_list (PANEL_KEY_BASE + "/general/applet_id_list", gconf.VALUE_STRING):
             if not self.applets.has_key (id):
-                self.applets[id] = PanelDelegate.PanelApplet (id, False)
+                dprint ("  %s", id);
+                self.applets[id] = PanelDelegate.PanelApplet (id)
+                
+        dprint ("Objects:");
         for id in self.client.get_list (PANEL_KEY_BASE + "/general/object_id_list", gconf.VALUE_STRING):
             if not self.objects.has_key (id):
-                self.objects[id] = PanelDelegate.PanelObject (id, False)
+                dprint ("  %s", id);
+                self.objects[id] = PanelDelegate.PanelObject (id)
 
     def __handle_id_list_change (self, change, dict, thing_class, added_class, removed_class):
         if not change.value or \
@@ -163,6 +187,9 @@ class PanelDelegate (userprofile.SourceDelegate):
                 dict[id].removed = True
                 removed.append (id)
 
+        dprint ("%s changed: (%s) added, (%s) removed\n",
+                change.key, added, removed);
+
         for id in added:
             self.source.emit ("changed", added_class (self.source, self, id))
         for id in removed:
@@ -240,6 +267,8 @@ class PanelDelegate (userprofile.SourceDelegate):
         if not thing.added:
             return
 
+        change.commit_change (mandatory)
+
         (client, address) = self.source.get_committing_client_and_address (mandatory)
 
         self.__copy_dir (self.client, client, address, PANEL_KEY_BASE + "/" + dir_name + "/" + thing.id)
@@ -258,6 +287,8 @@ class PanelDelegate (userprofile.SourceDelegate):
         if not thing.removed:
             return
 
+        change.commit_change (mandatory)
+
         (client, address) = self.source.get_committing_client_and_address (mandatory)
 
         id_list = self.__get_current_list (dict)
@@ -306,9 +337,12 @@ class PanelDelegate (userprofile.SourceDelegate):
                                           "object_id_list",
                                           "objects")
 
+    def get_path_description (self, path):
+        return "Panel launcher: %s"%os.path.basename(path)
+    
     def start_monitoring (self):
-        """Start monitoring for configuration changes."""
-        # Nothing to do here
+        # Need to do this here so gconf paths are setup
+        self.__read_panel_config ()
         pass
 
     def stop_monitoring (self):
@@ -320,13 +354,47 @@ class PanelDelegate (userprofile.SourceDelegate):
         # Nothing to do here
         pass
     
-    def apply (self):
+    def set_enforce_mandatory (self, enforce):
         # Nothing to do here
         pass
 
+    def __apply_foreach (self, source, path):
+        self.source.storage.extract (path, self.home_dir)
+
+    def apply (self, is_sabayon_session):
+        self.source.storage.foreach (self.__apply_foreach, source = self.name)
+        pass
+
 def get_gconf_delegate (source):
     return PanelDelegate (source)
 
+class PanelFileDelegate (userprofile.SourceDelegate):
+    def __init__ (self, source):
+        userprofile.SourceDelegate.__init__ (self, _("Panel File"), source, PANEL_LAUNCHER_DIR)
+
+    def handle_change (self, change):
+        dprint ("Ignoring file chage due to panel delegation: %s"%change)
+        return True
+
+    def start_monitoring (self):
+        pass
+
+    def stop_monitoring (self):
+        pass
+
+    def sync_changes (self):
+        pass
+
+    def set_enforce_mandatory (self, enforce):
+        pass
+
+    def apply (self, is_sabayon_session):
+        pass
+
+def get_files_delegate(source):
+    return PanelFileDelegate(source)
+
+
 #
 # Unit tests
 #
